"""Darker - apply black reformatting to only areas edited since the last commit"""

import logging
import sys
from difflib import unified_diff
from pathlib import Path
from typing import Generator, Iterable, List, Tuple

from darker.black_diff import BlackArgs, run_black
from darker.chooser import choose_lines
from darker.command_line import ISORT_INSTRUCTION, parse_command_line
from darker.config import dump_config
from darker.diff import diff_and_get_opcodes, opcodes_to_chunks
from darker.git import EditedLinenumsDiffer, git_get_modified_files
from darker.import_sorting import apply_isort, isort
from darker.linting import run_linter
from darker.utils import get_common_root, joinlines
from darker.verification import NotEquivalentError, verify_ast_unchanged

logger = logging.getLogger(__name__)


def format_edited_parts(
    srcs: Iterable[Path],
    revision: str,
    enable_isort: bool,
    linter_cmdlines: List[str],
    black_args: BlackArgs,
) -> Generator[Tuple[Path, str, str, List[str]], None, None]:
    """Black (and optional isort) formatting for chunks with edits since the last commit

    1. run isort on each edited file (optional)
    2. diff the given revision and worktree (optionally with isort modifications) for
       all file & dir paths on the command line
    3. extract line numbers in each edited to-file for changed lines
    4. run black on the contents of each edited to-file
    5. get a diff between the edited to-file and the reformatted content
    6. convert the diff into chunks, keeping original and reformatted content for each
       chunk
    7. choose reformatted content for each chunk if there were any changed lines inside
       the chunk in the edited to-file, or choose the chunk's original contents if no
       edits were done in that chunk
    8. concatenate all chosen chunks
    9. verify that the resulting reformatted source code parses to an identical AST as
       the original edited to-file
    10. write the reformatted source back to the original file
    11. run linter subprocesses for all edited files (11.-14. optional)
    12. diff the given revision and worktree (after isort and Black reformatting) for
        each file reported by a linter
    13. extract line numbers in each file reported by a linter for changed lines
    14. print only linter error lines which fall on changed lines

    :param srcs: Directories and files to re-format
    :param revision: The Git revision against which to compare the working tree
    :param enable_isort: ``True`` to also run ``isort`` first on each changed file
    :param linter_cmdlines: The command line(s) for running linters on the changed
                            files.
    :param black_args: Command-line arguments to send to ``black.FileMode``
    :return: A generator which yields details about changes for each file which should
             be reformatted, and skips unchanged files.

    """
    git_root = get_common_root(srcs)
    changed_files = git_get_modified_files(srcs, revision, git_root)
    edited_linenums_differ = EditedLinenumsDiffer(git_root, revision)

    for path_in_repo in sorted(changed_files):
        src = git_root / path_in_repo
        worktree_content = src.read_text()

        # 1. run isort
        if enable_isort:
            edited_content = apply_isort(
                worktree_content,
                src,
                black_args.get("config"),
                black_args.get("line_length"),
            )
        else:
            edited_content = worktree_content
        edited_lines = edited_content.splitlines()
        max_context_lines = len(edited_lines)
        for context_lines in range(max_context_lines + 1):
            # 2. diff the given revision and worktree for the file
            # 3. extract line numbers in the edited to-file for changed lines
            edited_linenums = edited_linenums_differ.revision_vs_lines(
                path_in_repo, edited_lines, context_lines
            )
            if (
                enable_isort
                and not edited_linenums
                and edited_content == worktree_content
            ):
                logger.debug("No changes in %s after isort", src)
                break

            # 4. run black
            formatted = run_black(src, edited_content, black_args)
            logger.debug("Read %s lines from edited file %s", len(edited_lines), src)
            logger.debug("Black reformat resulted in %s lines", len(formatted))

            # 5. get the diff between the edited and reformatted file
            opcodes = diff_and_get_opcodes(edited_lines, formatted)

            # 6. convert the diff into chunks
            black_chunks = list(opcodes_to_chunks(opcodes, edited_lines, formatted))

            # 7. choose reformatted content
            chosen_lines: List[str] = list(choose_lines(black_chunks, edited_linenums))

            # 8. concatenate chosen chunks
            result_str = joinlines(chosen_lines)

            # 9. verify
            logger.debug(
                "Verifying that the %s original edited lines and %s reformatted lines "
                "parse into an identical abstract syntax tree",
                len(edited_lines),
                len(chosen_lines),
            )
            try:
                verify_ast_unchanged(
                    edited_content, result_str, black_chunks, edited_linenums
                )
            except NotEquivalentError:
                # Diff produced misaligned chunks which couldn't be reconstructed into
                # a partially re-formatted Python file which produces an identical AST.
                # Try again with a larger `-U<context_lines>` option for `git diff`,
                # or give up if `context_lines` is already very large.
                if context_lines == max_context_lines:
                    raise
                logger.debug(
                    "AST verification failed. "
                    "Trying again with %s lines of context for `git diff -U`",
                    context_lines + 1,
                )
                continue
            else:
                # 10. A re-formatted Python file which produces an identical AST was
                #     created successfully - write an updated file or print the diff
                #     if there were any changes to the original
                if result_str != worktree_content:
                    # `result_str` is just `chosen_lines` concatenated with newlines.
                    # We need both forms when showing diffs or modifying files.
                    # Pass them both on to avoid back-and-forth conversion.
                    yield src, worktree_content, result_str, chosen_lines
                break
    # 11. run linter subprocesses for all edited files (11.-14. optional)
    # 12. diff the given revision and worktree (after isort and Black reformatting) for
    #     each file reported by a linter
    # 13. extract line numbers in each file reported by a linter for changed lines
    # 14. print only linter error lines which fall on changed lines
    for linter_cmdline in linter_cmdlines:
        run_linter(linter_cmdline, git_root, changed_files, revision)


def modify_file(path: Path, new_content: str) -> None:
    """Write new content to a file and inform the user by logging"""
    logger.info("Writing %s bytes into %s", len(new_content), path)
    path.write_text(new_content)


def print_diff(path: Path, old_content: str, new_lines: List[str]) -> None:
    """Print ``black --diff`` style output for the changes"""
    relative_path = path.resolve().relative_to(Path.cwd()).as_posix()
    diff = "\n".join(
        line.rstrip("\n")
        for line in unified_diff(
            old_content.splitlines(), new_lines, relative_path, relative_path,
        )
    )

    if sys.stdout.isatty():
        try:
            from pygments import highlight
            from pygments.formatters import TerminalFormatter
            from pygments.lexers import DiffLexer
        except ImportError:
            print(diff)
        else:
            print(highlight(diff, DiffLexer(), TerminalFormatter()))
    else:
        print(diff)


def main(argv: List[str] = None) -> int:
    """Parse the command line and apply black formatting for each source file

    :param argv: The command line arguments to the ``darker`` command
    :return: 1 if the ``--check`` argument was provided and at least one file was (or
             should be) reformatted; 0 otherwise.

    """
    if argv is None:
        argv = sys.argv[1:]
    args, config, config_nondefault = parse_command_line(argv)
    logging.basicConfig(level=args.log_level)
    if args.log_level == logging.INFO:
        formatter = logging.Formatter("%(levelname)s: %(message)s")
        logging.getLogger().handlers[0].setFormatter(formatter)

    # Make sure we don't get excessive debug log output from Black
    logging.getLogger("blib2to3.pgen2.driver").setLevel(logging.WARNING)

    if args.log_level <= logging.DEBUG:
        print("\n# Effective configuration:\n")
        print(dump_config(config))
        print("\n# Configuration options which differ from defaults:\n")
        print(dump_config(config_nondefault))
        print("\n")

    if args.isort and not isort:
        logger.error(f"{ISORT_INSTRUCTION} to use the `--isort` option.")
        exit(1)

    black_args = BlackArgs()
    if args.config:
        black_args["config"] = args.config
    if args.line_length:
        black_args["line_length"] = args.line_length
    if args.skip_string_normalization is not None:
        black_args["skip_string_normalization"] = args.skip_string_normalization

    paths = {Path(p) for p in args.src}
    some_files_changed = False
    # `new_content` is just `new_lines` concatenated with newlines.
    # We need both forms when showing diffs or modifying files.
    # Pass them both on to avoid back-and-forth conversion.
    for path, old_content, new_content, new_lines in format_edited_parts(
        paths, args.revision, args.isort, args.lint, black_args
    ):
        some_files_changed = True
        if args.diff:
            print_diff(path, old_content, new_lines)
        if not args.check and not args.diff:
            modify_file(path, new_content)
    return 1 if args.check and some_files_changed else 0


if __name__ == "__main__":
    RETVAL = main()
    sys.exit(RETVAL)



from argparse import ArgumentParser, Namespace
from typing import List, Tuple

from darker.argparse_helpers import LogLevelAction, NewlinePreservingFormatter
from darker.config import (
    DarkerConfig,
    get_effective_config,
    get_modified_config,
    load_config,
)
from darker.version import __version__

ISORT_INSTRUCTION = "Please run `pip install 'darker[isort]'`"


def make_argument_parser(require_src: bool) -> ArgumentParser:
    """Create the argument parser object

    :param require_src: ``True`` to require at least one path as a positional argument
                        on the command line. ``False`` to not require on.

    """
    description = [
        "Re-format Python source files by using",
        "- `isort` to sort Python import definitions alphabetically within logical"
        " sections",
        "- `black` to re-format code changed since the last Git commit",
    ]
    try:
        import isort
    except ImportError:
        isort = None
        description.extend(
            ["", f"{ISORT_INSTRUCTION} to enable sorting of import definitions"]
        )
    parser = ArgumentParser(
        description="\n".join(description), formatter_class=NewlinePreservingFormatter,
    )
    parser.register("action", "log_level", LogLevelAction)
    parser.add_argument(
        "src",
        nargs="+" if require_src else "*",
        help="Path(s) to the Python source file(s) to reformat",
        metavar="PATH",
    )
    parser.add_argument(
        "-r",
        "--revision",
        default="HEAD",
        help=(
            "Git revision against which to compare the working tree. Tags, branch"
            " names, commit hashes, and other expressions like HEAD~5 work here. Also"
            " a range like master...HEAD or master... can be used to compare the best"
            " common ancestor."
        ),
    )
    isort_help = ["Also sort imports using the `isort` package"]
    if not isort:
        isort_help.append(f". {ISORT_INSTRUCTION} to enable usage of this option.")
    parser.add_argument(
        "--diff",
        action="store_true",
        help=(
            "Don't write the files back, just output a diff for each file on stdout."
            " Highlight syntax on screen if the `pygments` package is available."
        ),
    )
    parser.add_argument(
        "--check",
        action="store_true",
        help=(
            "Don't write the files back, just return the status.  Return code 0 means"
            " nothing would change.  Return code 1 means some files would be"
            " reformatted."
        ),
    )
    parser.add_argument(
        "-i", "--isort", action="store_true", help="".join(isort_help),
    )
    parser.add_argument(
        "-L",
        "--lint",
        action="append",
        metavar="CMD",
        default=[],
        help=(
            "Also run a linter on changed files. CMD can be a name of path of the "
            "linter binary, or a full quoted command line"
        ),
    )
    parser.add_argument(
        "-c",
        "--config",
        metavar="PATH",
        help="Ask `black` and `isort` to read configuration from PATH.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="log_level",
        action="log_level",
        const=-10,
        help="Show steps taken and summarize modifications",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="log_level",
        action="log_level",
        const=10,
        help="Reduce amount of output",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="Show the version of `darker`",
    )
    parser.add_argument(
        "-S",
        "--skip-string-normalization",
        action="store_const",
        const=True,
        dest="skip_string_normalization",
        help="Don't normalize string quotes or prefixes",
    )
    parser.add_argument(
        "--no-skip-string-normalization",
        action="store_const",
        const=False,
        dest="skip_string_normalization",
        help=(
            "Normalize string quotes or prefixes. This can be used to override"
            " `skip_string_normalization = true` from a configuration file."
        ),
    )
    parser.add_argument(
        "-l",
        "--line-length",
        type=int,
        dest="line_length",
        help="How many characters per line to allow [default: 88]",
    )
    return parser


def parse_command_line(argv: List[str]) -> Tuple[Namespace, DarkerConfig, DarkerConfig]:
    """Return the parsed command line, using defaults from a configuration file

    Also return the effective configuration which combines defaults, the configuration
    read from ``pyproject.toml`` (or the path given in ``--config``), and command line
    arguments.

    Finally, also return the set of configuration options which differ from defaults.

    """
    # 1. Parse the paths of files/directories to process into `args.src`.
    parser_for_srcs = make_argument_parser(require_src=False)
    args = parser_for_srcs.parse_args(argv)

    # 2. Locate `pyproject.toml` based on those paths, or in the current directory if no
    #    paths were given. Load Darker configuration from it.
    config = load_config(args.src)

    # 3. Use configuration as defaults for re-parsing command line arguments, and don't
    #    require file/directory paths if they are specified in configuration.
    parser = make_argument_parser(require_src=not config.get("src"))
    parser.set_defaults(**config)
    args = parser.parse_args(argv)

    # 4. Also create a parser which uses the original default configuration values.
    #    This is used to find out differences between the effective configuration and
    #    default configuration values, and print them out in verbose mode.
    parser_with_original_defaults = make_argument_parser(require_src=True)
    return (
        args,
        get_effective_config(args),
        get_modified_config(parser_with_original_defaults, args),
    )



"""Configuration and fixtures for the Pytest based test suite"""

import sys
import types
from pathlib import Path
from subprocess import check_call
from typing import Dict, Optional
from unittest.mock import patch

import pytest
from black import find_project_root
from py.path import local as LocalPath

from darker.git import _git_check_output_lines


@pytest.fixture
def without_isort():
    with patch.dict(sys.modules, {"isort": None}):
        yield


@pytest.fixture
def with_isort():
    with patch.dict(sys.modules, {"isort": types.ModuleType("isort")}):
        yield


class GitRepoFixture:
    def __init__(self, root: LocalPath):
        self.root = root

    @classmethod
    def create_repository(cls, root: LocalPath) -> "GitRepoFixture":
        """Fixture method for creating a Git repository in the given directory"""
        check_call(["git", "init"], cwd=root)
        return cls(root)

    def _run(self, *args: str) -> None:
        """Helper method to run a Git command line in the repository root"""
        check_call(["git"] + list(args), cwd=self.root)

    def add(
        self, paths_and_contents: Dict[str, Optional[str]], commit: str = None
    ) -> Dict[str, LocalPath]:
        """Add/remove/modify files and optionally commit the changes

        :param paths_and_contents: Paths of the files relative to repository root, and
                                   new contents for the files as strings. ``None`` can
                                   be specified as the contents in order to remove a
                                   file.
        :param commit: The message for the commit, or ``None`` to skip making a commit.

        """
        absolute_paths = {
            relative_path: self.root / relative_path
            for relative_path in paths_and_contents
        }
        for relative_path, content in paths_and_contents.items():
            path = absolute_paths[relative_path]
            if content is None:
                self._run("rm", "--", relative_path)
            else:
                path.write(content, ensure=True)
                self._run("add", "--", relative_path)
        if commit:
            self._run("commit", "-m", commit)
        return absolute_paths

    def get_hash(self) -> str:
        """Return the commit hash at HEAD in the Git repository"""
        return _git_check_output_lines(["git", "rev-parse", "HEAD"], Path(self.root))[0]

    def create_branch(self, new_branch: str, start_point: str) -> None:
        """Fixture method to create and check out new branch at given starting point"""
        self._run("checkout", "-b", new_branch, start_point)


@pytest.fixture
def git_repo(tmpdir, monkeypatch):
    """Create a temporary Git repository and change current working directory into it"""
    repository = GitRepoFixture.create_repository(tmpdir)
    monkeypatch.chdir(tmpdir)
    return repository


@pytest.fixture
def find_project_root_cache_clear():
    """Clear LRU caching in :func:`black.find_project_root` before each test"""
    find_project_root.cache_clear()



# pylint: disable=too-many-arguments

"""Unit tests for :mod:`darker.linting`"""

from pathlib import Path

import pytest

from darker.linting import _parse_linter_line, run_linter


@pytest.mark.parametrize(
    "line, expect",
    [
        ("module.py:42: Description", (Path("module.py"), 42)),
        ("module.py:42:5: Description", (Path("module.py"), 42)),
        ("no-linenum.py: Description", (None, None)),
        ("mod.py:invalid-linenum:5: Description", (None, None)),
        ("invalid linter output", (None, None)),
    ],
)
def test_parse_linter_line(git_repo, monkeypatch, line, expect):
    """Linter output is parsed correctly"""
    monkeypatch.chdir(git_repo.root)
    result = _parse_linter_line(line, git_repo.root)
    assert result == expect


@pytest.mark.parametrize(
    "_descr, paths, location, expect",
    [
        ("No files to check, no output", [], "test.py:1:", []),
        (
            "Check one file, report on a modified line in test.py",
            ["one.py"],
            "test.py:1:",
            ["test.py:1: {git_repo.root}/one.py"],
        ),
        (
            "Check one file, report on a column of a modified line in test.py",
            ["one.py"],
            "test.py:1:42:",
            ["test.py:1:42: {git_repo.root}/one.py"],
        ),
        (
            "No output if report is on an unmodified line in test.py",
            ["one.py"],
            "test.py:2:42:",
            [],
        ),
        (
            "No output if report is on a column of an unmodified line in test.py",
            ["one.py"],
            "test.py:2:42:",
            [],
        ),
        (
            "Check two files, rpeort on a modified line in test.py",
            ["one.py", "two.py"],
            "test.py:1:",
            ["test.py:1: {git_repo.root}/one.py {git_repo.root}/two.py"],
        ),
        (
            "Check two files, rpeort on a column of a modified line in test.py",
            ["one.py", "two.py"],
            "test.py:1:42:",
            ["test.py:1:42: {git_repo.root}/one.py {git_repo.root}/two.py"],
        ),
        (
            "No output if 2-file report is on an unmodified line in test.py",
            ["one.py", "two.py"],
            "test.py:2:",
            [],
        ),
        (
            "No output if 2-file report is on a column of an unmodified line",
            ["one.py", "two.py"],
            "test.py:2:42:",
            [],
        ),
    ],
)
def test_run_linter(git_repo, monkeypatch, capsys, _descr, paths, location, expect):
    """Linter gets correct paths on command line and outputs just changed lines

    We use ``echo`` as our "linter". It just adds the paths of each file to lint as an
    "error" on a line of ``test.py``. What this test does is the equivalent of e.g.::

    - creating a ``test.py`` such that the first line is modified after the last commit
    - creating and committing ``one.py`` and ``two.py``
    - running::

          $ darker -L 'echo test.py:1:' one.py two.py
          test.py:1: git-repo-root/one.py git-repo-root/two.py

    """
    src_paths = git_repo.add({"test.py": "1\n2\n"}, commit="Initial commit")
    src_paths["test.py"].write("one\n2\n")
    monkeypatch.chdir(git_repo.root)
    cmdline = f"echo {location}"

    run_linter(cmdline, git_repo.root, {Path(p) for p in paths}, "HEAD")

    # We can now verify that the linter received the correct paths on its command line
    # by checking standard output from the our `echo` "linter".
    # The test cases also verify that only linter reports on modified lines are output.
    result = capsys.readouterr().out.splitlines()
    assert result == [line.format(git_repo=git_repo) for line in expect]



import re
from pathlib import Path
from textwrap import dedent
from unittest.mock import DEFAULT, Mock, call, patch

import pytest
import toml

from darker import black_diff
from darker.__main__ import main
from darker.command_line import make_argument_parser, parse_command_line
from darker.tests.helpers import filter_dict, raises_if_exception
from darker.utils import joinlines

pytestmark = pytest.mark.usefixtures("find_project_root_cache_clear")


@pytest.mark.parametrize("require_src, expect", [(False, []), (True, SystemExit)])
def test_make_argument_parser(require_src, expect):
    """Parser from ``make_argument_parser()`` fails if src required but not provided"""
    parser = make_argument_parser(require_src)
    with raises_if_exception(expect):

        args = parser.parse_args([])

        assert args.src == expect


@pytest.fixture
def darker_help_output(capsys):
    with pytest.raises(SystemExit):
        parse_command_line(["--help"])
    return re.sub(r'\s+', ' ', capsys.readouterr().out)


@pytest.mark.parametrize(
    "config, argv, expect",
    [
        (None, [], SystemExit),
        (
            None,
            ["file.py"],
            {"src": ["file.py"]},
        ),
        (
            {"src": ["file.py"]},
            [],
            {"src": ["file.py"]},
        ),
        (
            {"src": ["file.py"]},
            ["file.py"],
            {"src": ["file.py"]},
        ),
        (
            {"src": ["file1.py"]},
            ["file2.py"],
            {"src": ["file2.py"]},
        ),
    ],
)
def test_parse_command_line_config_src(
    tmpdir,
    monkeypatch,
    config,
    argv,
    expect,
):
    """The ``src`` positional argument from config and cmdline is handled correctly"""
    monkeypatch.chdir(tmpdir)
    if config is not None:
        toml.dump({"tool": {"darker": config}}, tmpdir / "pyproject.toml")
    with raises_if_exception(expect):

        args, effective_cfg, modified_cfg = parse_command_line(argv)

        assert filter_dict(args.__dict__, "src") == expect
        assert filter_dict(effective_cfg, "src") == expect
        assert filter_dict(modified_cfg, "src") == expect


@pytest.mark.parametrize(
    "argv, expect_value, expect_config, expect_modified",
    [
        (["."], ("src", ["."]), ("src", ["."]), ("src", ["."])),
        (
            ["."],
            ("revision", "HEAD"),
            ("revision", "HEAD"),
            ("revision", ...),
        ),
        (
            ["-rmaster", "."],
            ("revision", "master"),
            ("revision", "master"),
            ("revision", "master"),
        ),
        (
            ["--revision", "HEAD", "."],
            ("revision", "HEAD"),
            ("revision", "HEAD"),
            ("revision", ...),
        ),
        (["."], ("diff", False), ("diff", False), ("diff", ...)),
        (["--diff", "."], ("diff", True), ("diff", True), ("diff", True)),
        (["."], ("check", False), ("check", False), ("check", ...)),
        (["--check", "."], ("check", True), ("check", True), ("check", True)),
        (["."], ("isort", False), ("isort", False), ("isort", ...)),
        (["-i", "."], ("isort", True), ("isort", True), ("isort", True)),
        (["--isort", "."], ("isort", True), ("isort", True), ("isort", True)),
        (["."], ("lint", []), ("lint", []), ("lint", ...)),
        (
            ["-L", "pylint", "."],
            ("lint", ["pylint"]),
            ("lint", ["pylint"]),
            ("lint", ["pylint"]),
        ),
        (
            ["--lint", "flake8", "-L", "mypy", "."],
            ("lint", ["flake8", "mypy"]),
            ("lint", ["flake8", "mypy"]),
            ("lint", ["flake8", "mypy"]),
        ),
        (["."], ("config", None), ("config", None), ("config", ...)),
        (
            ["-c", "my.cfg", "."],
            ("config", "my.cfg"),
            ("config", "my.cfg"),
            ("config", "my.cfg"),
        ),
        (
            ["--config=my.cfg", "."],
            ("config", "my.cfg"),
            ("config", "my.cfg"),
            ("config", "my.cfg"),
        ),
        (["."], ("log_level", 30), ("log_level", "WARNING"), ("log_level", ...)),
        (
            ["-v", "."],
            ("log_level", 20),
            ("log_level", "INFO"),
            ("log_level", "INFO"),
        ),
        (
            ["--verbose", "-v", "."],
            ("log_level", 10),
            ("log_level", "DEBUG"),
            ("log_level", "DEBUG"),
        ),
        (
            ["-q", "."],
            ("log_level", 40),
            ("log_level", "ERROR"),
            ("log_level", "ERROR"),
        ),
        (
            ["--quiet", "-q", "."],
            ("log_level", 50),
            ("log_level", "CRITICAL"),
            ("log_level", "CRITICAL"),
        ),
        (
            ["."],
            ("skip_string_normalization", None),
            ("skip_string_normalization", None),
            ("skip_string_normalization", ...),
        ),
        (
            ["-S", "."],
            ("skip_string_normalization", True),
            ("skip_string_normalization", True),
            ("skip_string_normalization", True),
        ),
        (
            ["--skip-string-normalization", "."],
            ("skip_string_normalization", True),
            ("skip_string_normalization", True),
            ("skip_string_normalization", True),
        ),
        (
            ["--no-skip-string-normalization", "."],
            ("skip_string_normalization", False),
            ("skip_string_normalization", False),
            ("skip_string_normalization", False),
        ),
        (
            ["."],
            ("line_length", None),
            ("line_length", None),
            ("line_length", ...),
        ),
        (
            ["-l=88", "."],
            ("line_length", 88),
            ("line_length", 88),
            ("line_length", 88),
        ),
        (
            ["--line-length", "99", "."],
            ("line_length", 99),
            ("line_length", 99),
            ("line_length", 99),
        ),
    ],
)
def test_parse_command_line(
    tmpdir, monkeypatch, argv, expect_value, expect_config, expect_modified
):
    monkeypatch.chdir(tmpdir)
    args, effective_cfg, modified_cfg = parse_command_line(argv)

    arg_name, expect_arg_value = expect_value
    assert getattr(args, arg_name) == expect_arg_value

    option, expect_config_value = expect_config
    if expect_config_value is ...:
        assert option not in effective_cfg
    else:
        assert effective_cfg[option] == expect_config_value

    modified_option, expect_modified_value = expect_modified
    if expect_modified_value is ...:
        assert modified_option not in modified_cfg
    else:
        assert modified_cfg[modified_option] == expect_modified_value


def test_help_description_without_isort_package(without_isort, darker_help_output):
    assert (
        "Please run `pip install 'darker[isort]'` to enable sorting of import "
        "definitions" in darker_help_output
    )


def test_help_isort_option_without_isort_package(without_isort, darker_help_output):
    assert (
        "Please run `pip install 'darker[isort]'` to enable usage of this option."
        in darker_help_output
    )


def test_help_with_isort_package(with_isort, darker_help_output):
    assert "Please run" not in darker_help_output


@pytest.mark.parametrize(
    "options, expect",
    [
        ([], call()),
        (['-c', 'black.cfg'], call(line_length=81, string_normalization=True)),
        (['--config', 'black.cfg'], call(line_length=81, string_normalization=True)),
        (['-S'], call(string_normalization=False)),
        (['--skip-string-normalization'], call(string_normalization=False)),
        (['-l', '90'], call(line_length=90)),
        (['--line-length', '90'], call(line_length=90)),
        (['-c', 'black.cfg', '-S'], call(line_length=81, string_normalization=False)),
        (
            ['-c', 'black.cfg', '-l', '90'],
            call(line_length=90, string_normalization=True),
        ),
        (['-l', '90', '-S'], call(line_length=90, string_normalization=False)),
        (
            ['-c', 'black.cfg', '-l', '90', '-S'],
            call(line_length=90, string_normalization=False),
        ),
    ],
)
def test_black_options(monkeypatch, tmpdir, git_repo, options, expect):
    monkeypatch.chdir(tmpdir)
    (tmpdir / 'pyproject.toml').write("[tool.black]\n")
    (tmpdir / 'black.cfg').write(
        dedent(
            """
            [tool.black]
            line-length = 81
            skip-string-normalization = false
            """
        )
    )
    added_files = git_repo.add(
        {"main.py": 'print("Hello World!")\n'}, commit="Initial commit"
    )
    added_files["main.py"].write('print ("Hello World!")\n')
    with patch.object(black_diff, 'Mode', wraps=black_diff.Mode) as Mode:

        main(options + [str(path) for path in added_files.values()])

    _, expect_args, expect_kwargs = expect
    Mode.assert_called_once_with(*expect_args, **expect_kwargs)


@pytest.mark.parametrize(
    "config, options, expect",
    [
        ([], [], call()),
        ([], ['--skip-string-normalization'], call(string_normalization=False)),
        ([], ['--no-skip-string-normalization'], call(string_normalization=True)),
        (['skip_string_normalization = false'], [], call(string_normalization=True)),
        (
            ['skip_string_normalization = false'],
            ['--skip-string-normalization'],
            call(string_normalization=False),
        ),
        (
            ['skip_string_normalization = false'],
            ['--no-skip-string-normalization'],
            call(string_normalization=True),
        ),
        (['skip_string_normalization = true'], [], call(string_normalization=False)),
        (
            ['skip_string_normalization = true'],
            ['--skip-string-normalization'],
            call(string_normalization=False),
        ),
        (
            ['skip_string_normalization = true'],
            ['--no-skip-string-normalization'],
            call(string_normalization=True),
        ),
    ],
)
def test_black_options_skip_string_normalization(git_repo, config, options, expect):
    """Black string normalization config and cmdline option are combined correctly"""
    added_files = git_repo.add(
        {"main.py": "foo", "pyproject.toml": joinlines(["[tool.black]"] + config)},
        commit="Initial commit",
    )
    added_files["main.py"].write("bar")
    mode_class_mock = Mock(wraps=black_diff.Mode)
    # Speed up tests by mocking `format_str` to skip running Black
    format_str = Mock(return_value="bar")
    with patch.multiple(black_diff, Mode=mode_class_mock, format_str=format_str):

        main(options + [str(path) for path in added_files.values()])

    assert mode_class_mock.call_args_list == [expect]


@pytest.mark.parametrize(
    'options, expect',
    [
        (["a.py"], ({Path("a.py")}, "HEAD", False, [], {})),
        (["--isort", "a.py"], ({Path("a.py")}, "HEAD", True, [], {})),
        (
            ["--config", "my.cfg", "a.py"],
            ({Path("a.py")}, "HEAD", False, [], {"config": "my.cfg"}),
        ),
        (
            ["--line-length", "90", "a.py"],
            ({Path("a.py")}, "HEAD", False, [], {"line_length": 90}),
        ),
        (
            ["--skip-string-normalization", "a.py"],
            ({Path("a.py")}, "HEAD", False, [], {"skip_string_normalization": True}),
        ),
        (["--diff", "a.py"], ({Path("a.py")}, "HEAD", False, [], {})),
    ],
)
def test_options(tmpdir, monkeypatch, options, expect):
    """The main engine is called with correct parameters based on the command line

    Executed in a clean directory so Darker's own ``pyproject.toml`` doesn't interfere.

    """
    monkeypatch.chdir(tmpdir)
    (tmpdir / "my.cfg").write("")
    with patch('darker.__main__.format_edited_parts') as format_edited_parts:

        retval = main(options)

    format_edited_parts.assert_called_once_with(*expect)
    assert retval == 0


@pytest.mark.parametrize(
    'check, changes, expect_retval',
    [(False, False, 0), (False, True, 0), (True, False, 0), (True, True, 1)],
)
def test_main_retval(check, changes, expect_retval):
    """main() return value is correct based on --check and the need to reformat files"""
    format_edited_parts = Mock()
    format_edited_parts.return_value = (
        [(Path('/dummy.py'), 'old\n', 'new\n', ['new'])] if changes else []
    )
    check_arg_maybe = ['--check'] if check else []
    with patch.multiple(
        'darker.__main__', format_edited_parts=format_edited_parts, modify_file=DEFAULT
    ):

        retval = main(check_arg_maybe + ['a.py'])

    assert retval == expect_retval



from pathlib import Path

import pytest

from darker.black_diff import BlackArgs, read_black_config, run_black


@pytest.mark.parametrize(
    "config_path, config_lines, expect",
    [
        (None, ['line-length = 79'], {'line_length': 79}),
        ("custom.toml", ['line-length = 99'], {'line_length': 99}),
        (
            "custom.toml",
            ['skip-string-normalization = true'],
            {'skip_string_normalization': True},
        ),
        (
            "custom.toml",
            ['skip-string-normalization = false'],
            {'skip_string_normalization': False},
        ),
        ("custom.toml", ["target-version = ['py37']"], {}),
        ("custom.toml", ["include = '\\.pyi$'"], {}),
        ("custom.toml", ["exclude = '\\.pyx$'"], {}),
    ],
)
def test_black_config(tmpdir, config_path, config_lines, expect):
    tmpdir = Path(tmpdir)
    src = tmpdir / "src.py"
    toml = tmpdir / (config_path or "pyproject.toml")

    toml.write_text("[tool.black]\n{}\n".format('\n'.join(config_lines)))

    config = read_black_config(src, config_path and str(toml))
    assert config == expect


def test_run_black(tmpdir):
    src_contents = "print ( '42' )\n"

    result = run_black(Path(tmpdir / "src.py"), src_contents, BlackArgs())

    assert result == ['print("42")']



"""Helpers for listing modified files and getting unmodified content from Git"""

import logging
import re
import sys
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from subprocess import CalledProcessError, check_output
from typing import Iterable, List, Set

from darker.diff import diff_and_get_opcodes, opcodes_to_edit_linenums

logger = logging.getLogger(__name__)


COMMIT_RANGE_SPLIT_RE = re.compile(r"\.{2,3}")
# Split a revision range into the "from" and "to" revisions and the dots in between.
# Handles these cases:
# <rev>..   <rev>..<rev>   ..<rev>
# <rev>...  <rev>...<rev>  ...<rev>
COMMIT_RANGE_RE = re.compile(r"(.*?)(\.{2,3})(.*)$")


# A colon is an invalid character in tag/branch names. Use that in the special value for
# denoting the working tree as one of the "revisions" in revision ranges.
WORKTREE = ":WORKTREE:"


def git_get_unmodified_content(path: Path, revision: str, cwd: Path) -> List[str]:
    """Get unmodified text lines of a file at a Git revision

    :param path: The relative path of the file in the Git repository
    :param revision: The Git revision for which to get the file content. This can also
                     be a range of commits like ``master..HEAD`` or ``master...HEAD``,
                     in which case the leftmost end of the range is used.
    :param cwd: The root of the Git repository

    """
    commit = COMMIT_RANGE_SPLIT_RE.split(revision, 1)[0]
    cmd = ["git", "show", f"{commit}:./{path}"]
    logger.debug("[%s]$ %s", cwd, " ".join(cmd))
    try:
        return check_output(cmd, cwd=str(cwd), encoding="utf-8").splitlines()
    except CalledProcessError as exc_info:
        if exc_info.returncode == 128:
            # The file didn't exist at the given revision. Act as if it was an empty
            # file, so all current lines appear as edited.
            return []
        else:
            raise


@dataclass(frozen=True)
class RevisionRange:
    """Represent a range of commits in a Git repository for comparing differences

    ``rev1`` is the "old" revision, and ``rev2``, the "new" revision which should be
    compared against ``rev1``.

    If ``use_common_ancestor`` is true, the comparison should not be made against
    ``rev1`` but instead against the branch point, i.e. the latest commit which is
    common to both ``rev1`` and ``rev2``. This is useful e.g. when CI is doing a check
    on a feature branch, and there have been commits in the main branch after the branch
    point. Without the ability to compare to the branch point, Darker would suggest
    corrections to formatting on lines changes in the main branch even if those lines
    haven't been touched in the feature branch.

    """

    rev1: str
    rev2: str = WORKTREE
    use_common_ancestor: bool = False

    def __post_init__(self) -> None:
        if self.rev2 == "":
            super().__setattr__("rev2", WORKTREE)

    @classmethod
    def parse(cls, revision_range: str) -> "RevisionRange":
        """Convert a range expression to a ``RevisionRange`' object

        >>> RevisionRange.parse("a..b")
        RevisionRange(rev1='a', rev2='b', use_common_ancestor=False)
        >>> RevisionRange.parse("a...b")
        RevisionRange(rev1='a', rev2='b', use_common_ancestor=True)
        >>> RevisionRange.parse("a..")
        RevisionRange(rev1='a', rev2=':WORKTREE:', use_common_ancestor=False)
        >>> RevisionRange.parse("a...")
        RevisionRange(rev1='a', rev2=':WORKTREE:', use_common_ancestor=True)

        """
        match = COMMIT_RANGE_RE.match(revision_range)
        if match:
            rev1, range_dots, rev2 = match.groups()
            use_common_ancestor = range_dots == "..."
            return cls(rev1 or "HEAD", rev2, use_common_ancestor)
        return cls(
            revision_range or "HEAD",
            WORKTREE,
            use_common_ancestor=revision_range not in ["", "HEAD"],
        )


def should_reformat_file(path: Path) -> bool:
    return path.exists() and path.suffix == ".py"


def _git_check_output_lines(cmd: List[str], cwd: Path) -> List[str]:
    """Log command line, run Git, split stdout to lines, exit with 123 on error"""
    logger.debug("[%s]$ %s", cwd, " ".join(cmd))
    try:
        return check_output(cmd, cwd=str(cwd)).decode("utf-8").splitlines()
    except CalledProcessError as exc_info:
        if exc_info.returncode == 128:
            # Bad revision or another Git failure
            sys.exit(123)
        else:
            raise


def git_get_modified_files(
    paths: Iterable[Path], revision: str, cwd: Path
) -> Set[Path]:
    """Ask Git for modified and untracked files

    - ``git diff --name-only --relative <rev> -- <path(s)>``
    - ``git ls-files --others --exclude-standard -- <path(s)>``

    Return file names relative to the Git repository root.

    :paths: Paths to the files to diff
    :param revision: Git revision to compare current working tree against
    :cwd: The Git repository root

    """
    relative_paths = {p.resolve().relative_to(cwd) for p in paths}
    str_paths = [str(path) for path in relative_paths]
    diff_cmd = [
        "git",
        "diff",
        "--name-only",
        "--relative",
        # `revision` is inserted here if non-empty
        "--",
        *str_paths,
    ]
    if revision:
        diff_cmd.insert(diff_cmd.index("--"), revision)
    lines = _git_check_output_lines(diff_cmd, cwd)
    ls_files_cmd = [
        "git",
        "ls-files",
        "--others",
        "--exclude-standard",
        "--",
        *str_paths,
    ]
    lines.extend(_git_check_output_lines(ls_files_cmd, cwd))
    changed_paths = (Path(line) for line in lines)
    return {path for path in changed_paths if should_reformat_file(cwd / path)}


@dataclass(frozen=True)
class EditedLinenumsDiffer:
    """Find out changed lines for a file compared to a given Git revision"""

    git_root: Path
    revision: str

    @lru_cache(maxsize=1)
    def revision_vs_worktree(self, path_in_repo: Path, context_lines: int) -> List[int]:
        """Return numbers of lines changed between a given revision and the worktree"""
        lines = (self.git_root / path_in_repo).read_text("utf-8").splitlines()
        return self.revision_vs_lines(path_in_repo, lines, context_lines)

    def revision_vs_lines(
        self, path_in_repo: Path, lines: List[str], context_lines: int
    ) -> List[int]:
        """For file `path_in_repo`, return changed line numbers from given revision

        :param path_in_repo: Path of the file to compare, relative to repository root
        :param lines: The contents to compare to, e.g. from current working tree
        :return: Line numbers of lines changed between the revision and given content

        """
        revision_lines = git_get_unmodified_content(
            path_in_repo, self.revision, self.git_root
        )
        edited_opcodes = diff_and_get_opcodes(revision_lines, lines)
        return list(opcodes_to_edit_linenums(edited_opcodes, context_lines))



"""Helper functions for unit tests"""

import sys
from typing import Any, ContextManager, Dict, Union

import pytest
from _pytest.python_api import RaisesContext

if sys.version_info >= (3, 7):
    from contextlib import nullcontext
else:
    from contextlib import suppress as nullcontext


def filter_dict(dct: Dict[str, Any], filter_key: str) -> Dict[str, Any]:
    """Return only given keys with their values from a dictionary"""
    return {key: value for key, value in dct.items() if key == filter_key}


def raises_if_exception(expect: Any) -> Union[RaisesContext[Any], ContextManager[None]]:
    """Return a ``pytest.raises`` context manager only if expecting an exception

    If the expected value is not an exception, return a dummy context manager.
    On Python 3.8+ :class:`contextlib.nullcontext` is returned,
    while on older Pythons :class:`contextlib.suppress` is used instead.

    """
    if isinstance(expect, type) and issubclass(expect, BaseException):
        return pytest.raises(expect)
    else:
        return nullcontext()



"""Diff text and get line numbers of changes or chunks of original and changed content

The functions in this module implement

- diffing text files, returning opcodes
- turning opcodes into a list of line numbers of changed lines
- turning opcodes into chunks of original and modified text

In our case, we run a diff between original and user-edited source code.
Another diff is done between user-edited and Black-reformatted source code
as returned by :func:`black.black_diff.run_black_for_content`.

    >>> src_lines = [
    ...     'for i in range(5): print(i)',
    ...     'print("done")'
    ... ]

    >>> dst_lines = [
    ...     'for i in range(5):',
    ...     '    print(i)',
    ...     'print("done")'
    ... ]

:func:`diff_and_get_opcodes`.
divides a diff between the original and reformatted content
into alternating chunks of
intact (represented by the 'equal' tag) and
modified ('delete', 'replace' or 'insert' tag) lines.
Each chunk is an opcode represented by the tag and the corresponding 0-based line ranges
in the original and reformatted content, e.g.::

    >>> opcodes = diff_and_get_opcodes(src_lines, dst_lines)
    >>> len(opcodes)
    2
    >>> opcodes[0]  # split 'for' loop into two lines
    ('replace', 0, 1, 0, 2)
    >>> opcodes[1]  # keep 'print("done")' as such
    ('equal', 1, 2, 2, 3)

:func:`opcodes_to_chunks` picks the lines
from original and reformatted content for each opcode.
It combines line content with the 1-based line offset in the original content, e.g.::

    >>> chunks = list(opcodes_to_chunks(opcodes, src_lines, dst_lines))
    >>> len(chunks)
    2
    >>> chunks[0]  # (<offset in orig content>, <original lines>, <reformatted lines>)
    (1,
     ['for i in range(5): print(i)'],
     ['for i in range(5):',
      '    print(i)'])
    >>> chunks[1]
    (2,
     ['print("done")'],
     ['print("done")'])

By concatenating the second items in these tuples, i.e. original lines,
the original file can be reconstructed.

By concatenating the third items, i.e. reformatted lines,
the complete output from Black can be reconstructed.

By concatenating and choosing either the second or third item,
a mixed result with only selected regions reformatted can be reconstructed.

"""

import logging
from difflib import SequenceMatcher
from typing import Generator, List, Tuple

logger = logging.getLogger(__name__)


def diff_and_get_opcodes(
    src_lines: List[str], dst_lines: List[str]
) -> List[Tuple[str, int, int, int, int]]:
    """Return opcodes and line numbers for chunks in the diff of two lists of strings

    The opcodes are 5-tuples for each chunk with

    - the tag of the operation ('equal', 'delete', 'replace' or 'insert')
    - the number of the first line in the chunk in the from-file
    - the number of the last line in the chunk in the from-file
    - the number of the first line in the chunk in the to-file
    - the number of the last line in the chunk in the to-file

    Line numbers are zero based.

    """
    matcher = SequenceMatcher(None, src_lines, dst_lines, autojunk=False)
    opcodes = matcher.get_opcodes()
    logger.debug(
        "Diff between edited and reformatted has %s opcode%s",
        len(opcodes),
        "s" if len(opcodes) > 1 else "",
    )
    return opcodes


def _validate_opcodes(opcodes: List[Tuple[str, int, int, int, int]]) -> None:
    """Make sure every other opcode is an 'equal' tag"""
    assert all(
        (tag1 == "equal") != (tag2 == "equal")
        for (tag1, _, _, _, _), (tag2, _, _, _, _) in zip(opcodes[:-1], opcodes[1:])
    ), opcodes


def opcodes_to_edit_linenums(
    opcodes: List[Tuple[str, int, int, int, int]], context_lines: int
) -> Generator[int, None, None]:
    """Convert diff opcodes to line numbers of edited lines in the destination file

    :param opcodes: The diff opcodes to convert
    :param context_lines: The number of lines before and after an edited line to mark
                          edited as well

    """
    if not opcodes:
        return
    _validate_opcodes(opcodes)
    prev_chunk_end = 1
    _tag, _i1, _i2, _j1, end = opcodes[-1]
    for tag, _i1, _i2, j1, j2 in opcodes:
        if tag != "equal":
            chunk_end = min(j2 + 1 + context_lines, end + 1)
            yield from range(max(j1 + 1 - context_lines, prev_chunk_end), chunk_end)
            prev_chunk_end = chunk_end


def opcodes_to_chunks(
    opcodes: List[Tuple[str, int, int, int, int]],
    src_lines: List[str],
    dst_lines: List[str],
) -> Generator[Tuple[int, List[str], List[str]], None, None]:
    """Convert each diff opcode to a line number and original plus modified lines

    Each chunk is a 3-tuple with

    - the 1-based number of the first line in the chunk in the from-file
    - the original lines of the chunk in the from-file
    - the modified lines of the chunk in the to-file

    Based on this, the patch can be constructed by choosing either original or modified
    lines for each chunk and concatenating them together.

    """
    _validate_opcodes(opcodes)
    for tag, i1, i2, j1, j2 in opcodes:
        yield i1 + 1, src_lines[i1:i2], dst_lines[j1:j2]



"""Re-format Python source code using Black

In examples below, a simple two-line snippet is used.
The first line will be reformatted by Black, and the second left intact::

    >>> from unittest.mock import Mock
    >>> src = Mock()
    >>> src_content = '''\\
    ... for i in range(5): print(i)
    ... print("done")
    ... '''

First, :func:`run_black` uses Black to reformat the contents of a given file.
Reformatted lines are returned e.g.::

    >>> dst_lines = run_black(src, src_content, black_args={})
    >>> dst_lines
    ['for i in range(5):',
     '    print(i)',
     'print("done")']

See :mod:`darker.diff` and :mod:`darker.chooser`
for how this result is further processed with:

- :func:`~darker.diff.diff_and_get_opcodes`
  to get a diff of the reformatting
- :func:`~darker.diff.opcodes_to_chunks`
  to split the diff into chunks of original and reformatted content
- :func:`~darker.chooser.choose_lines`
  to reconstruct the source code from original and reformatted chunks
  based on whether reformats touch user-edited lines

"""

import logging
import sys
from functools import lru_cache
from pathlib import Path
from typing import List, Optional, Set, cast

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

# `FileMode as Mode` required to satisfy mypy==0.782. Strange.
from black import FileMode as Mode
from black import TargetVersion, find_pyproject_toml, format_str, parse_pyproject_toml

logger = logging.getLogger(__name__)


class BlackArgs(TypedDict, total=False):
    config: str
    line_length: int
    skip_string_normalization: bool


class BlackModeAttributes(TypedDict, total=False):
    target_versions: Set[TargetVersion]
    line_length: int
    string_normalization: bool
    is_pyi: bool


@lru_cache(maxsize=1)
def read_black_config(src: Path, value: Optional[str]) -> BlackArgs:
    """Read the black configuration from pyproject.toml"""
    value = value or find_pyproject_toml((str(src),))

    if not value:
        return BlackArgs()

    config = parse_pyproject_toml(value)

    return cast(
        BlackArgs,
        {
            key: value
            for key, value in config.items()
            if key in ["line_length", "skip_string_normalization"]
        },
    )


def run_black(src: Path, src_contents: str, black_args: BlackArgs) -> List[str]:
    """Run the black formatter for the Python source code given as a string

    Return lines of the original file as well as the formatted content.

    :param src: The originating file path for the source code
    :param src_contents: The source code as a string
    :param black_args: Command-line arguments to send to ``black.FileMode``

    """
    config = black_args.pop("config", None)
    combined_args = read_black_config(src, config)
    combined_args.update(black_args)

    effective_args = BlackModeAttributes()
    if "line_length" in combined_args:
        effective_args["line_length"] = combined_args["line_length"]
    if "skip_string_normalization" in combined_args:
        # The ``black`` command line argument is
        # ``--skip-string-normalization``, but the parameter for
        # ``black.Mode`` needs to be the opposite boolean of
        # ``skip-string-normalization``, hence the inverse boolean
        effective_args["string_normalization"] = not combined_args[
            "skip_string_normalization"
        ]

    # Override defaults and pyproject.toml settings if they've been specified
    # from the command line arguments
    mode = Mode(**effective_args)

    dst_contents = format_str(src_contents, mode=mode)
    dst_lines: List[str] = dst_contents.splitlines()
    return dst_lines



"""Picking of original or reformatted chunks based on edits

The :func:`choose_lines` function should be fed with
a list of line numbers which were edited after the last commit,
plus output from :func:`darker.black_diff.opcodes_to_chunks`.
It reconstructs the Python source code file from chunks
while choosing either the original or reformatted version of each chunk.
The original is chosen if no edited line number falls on original chunk lines,
and the reformatted if an edit was seen on any of the lines in the chunk.

Example::

    >>> reconstruction = choose_lines(
    ...     [ ( 1,                        # chunk starts on line 1 in original content
    ...         ['original line 1',
    ...          'original line 2'],
    ...         ['reformatted lines 1-2']
    ...       ),
    ...       ( 3,                        # chunk starts on line 3 in original content
    ...         ['original line 3',
    ...          'original line 4'],
    ...         ['reformatted lines 3-4']
    ...       )
    ...      ],
    ...      [2]                          # only line 2 was edited since last commit
    ... )
    >>> list(reconstruction)
    ['reformatted lines 1-2', 'original line 3', 'original line 4']

"""

import logging
from typing import Generator, Iterable, List, Tuple

logger = logging.getLogger(__name__)


def _any_item_in_range(items: List[int], start: int, length: int) -> bool:
    """Return ``True`` if any item falls inside the slice ``[start : start + length]``

    If ``length == 0``, add one to make sure an edit at the position of an inserted
    chunk causes the reformatted version to be chosen for that chunk.

    """
    end = start + (length or 1) - 1
    has_edits = any(start <= n <= end for n in items)
    line_range = f'line {start}' if end == start else f'lines {start}-{end}'
    if has_edits:
        logger.debug("Found edits on %s", line_range)
    else:
        logger.debug("Found no edits on %s", line_range)
    return has_edits


def choose_lines(
    black_chunks: Iterable[Tuple[int, List[str], List[str]]], edit_linenums: List[int],
) -> Generator[str, None, None]:
    """Choose formatted chunks for edited areas, original chunks for non-edited"""
    for original_lines_offset, original_lines, formatted_lines in black_chunks:
        chunk_has_edits = _any_item_in_range(
            edit_linenums, original_lines_offset, len(original_lines)
        )
        if chunk_has_edits:
            choice = (
                "unmodified" if formatted_lines == original_lines else "reformatted"
            )
            chosen_lines = formatted_lines
        else:
            choice = 'original'
            chosen_lines = original_lines
        logger.debug(
            'Using %s %s %s at line %s',
            len(chosen_lines),
            choice,
            'line' if len(chosen_lines) == 1 else 'lines',
            original_lines_offset,
        )
        yield from chosen_lines



from pathlib import Path
from textwrap import dedent

from darker.utils import debug_dump, get_common_root, get_path_ancestry, joinlines


def test_debug_dump(capsys):
    debug_dump([(1, ["black"], ["chunks"])], "old content", "new content", [2, 3])
    assert capsys.readouterr().out == (
        dedent(
            """\
            --------------------------------------------------------------------------------
             -   1 black
             +     chunks
            --------------------------------------------------------------------------------
            """
        )
    )


def test_joinlines():
    result = joinlines(["a", "b", "c"])
    assert result == "a\nb\nc\n"


def test_get_common_root(tmpdir):
    tmpdir = Path(tmpdir)
    path1 = tmpdir / "a" / "b" / "c" / "d"
    path2 = tmpdir / "a" / "e" / ".." / "b" / "f" / "g"
    path3 = tmpdir / "a" / "h" / ".." / "b" / "i"
    result = get_common_root([path1, path2, path3])
    assert result == tmpdir / "a" / "b"


def test_get_common_root_of_directory(tmpdir):
    tmpdir = Path(tmpdir)
    result = get_common_root([tmpdir])
    assert result == tmpdir


def test_get_path_ancestry_for_directory(tmpdir):
    tmpdir = Path(tmpdir)
    result = list(get_path_ancestry(tmpdir))
    assert result[-1] == tmpdir
    assert result[-2] == tmpdir.parent


def test_get_path_ancestry_for_file(tmpdir):
    tmpdir = Path(tmpdir)
    dummy = tmpdir / "dummy"
    dummy.write_text("dummy")
    result = list(get_path_ancestry(dummy))
    assert result[-1] == tmpdir
    assert result[-2] == tmpdir.parent



"""Verification for unchanged AST before and after reformatting"""

from typing import List, Tuple

from black import assert_equivalent

from darker.utils import debug_dump


class NotEquivalentError(Exception):
    pass


def verify_ast_unchanged(
    edited_to_file_str: str,
    reformatted_str: str,
    black_chunks: List[Tuple[int, List[str], List[str]]],
    edited_linenums: List[int],
) -> None:
    """Verify that source code parses to the same AST before and after reformat"""
    try:
        assert_equivalent(edited_to_file_str, reformatted_str)
    except AssertionError as exc_info:
        debug_dump(black_chunks, edited_to_file_str, reformatted_str, edited_linenums)
        raise NotEquivalentError(str(exc_info))



"""Tests for the ``darker.argparse_helpers`` module"""

from argparse import ArgumentParser, Namespace
from logging import CRITICAL, DEBUG, ERROR, INFO, NOTSET, WARNING

import pytest

from darker import argparse_helpers
from darker.tests.helpers import raises_if_exception


@pytest.mark.parametrize(
    "line, width, indent, expect",
    [
        ("", 0, "    ", ValueError),
        ("", 1, "    ", []),
        (
            "lorem ipsum dolor sit amet",
            9,
            "    ",
            ["    lorem", "    ipsum", "    dolor", "    sit", "    amet"],
        ),
        (
            "lorem ipsum dolor sit amet",
            15,
            "    ",
            ["    lorem ipsum", "    dolor sit", "    amet"],
        ),
    ],
)
def test_fill_line(line, width, indent, expect):
    """``_fill_line()`` wraps lines correctly"""
    with raises_if_exception(expect):

        result = argparse_helpers._fill_line(  # pylint: disable=protected-access
            line, width, indent
        )

        assert result.splitlines() == expect


@pytest.mark.parametrize(
    "text, width, indent, expect",
    [
        (
            "lorem ipsum dolor sit amet",
            15,
            "    ",
            ["    lorem ipsum", "    dolor sit", "    amet"],
        ),
        (
            "lorem\nipsum dolor sit amet",
            15,
            "    ",
            ["    lorem", "    ipsum dolor", "    sit amet"],
        ),
    ],
)
def test_newline_preserving_formatter(text, width, indent, expect):
    """``NewlinePreservingFormatter`` wraps lines and keeps newlines correctly"""
    formatter = argparse_helpers.NewlinePreservingFormatter("dummy")

    result = formatter._fill_text(  # pylint: disable=protected-access
        text, width, indent
    )

    assert result.splitlines() == expect


@pytest.mark.parametrize(
    "const, initial, expect",
    [
        (10, NOTSET, DEBUG),
        (10, DEBUG, INFO),
        (10, INFO, WARNING),
        (10, WARNING, ERROR),
        (10, ERROR, CRITICAL),
        (10, CRITICAL, CRITICAL),
        (-10, DEBUG, DEBUG),
        (-10, INFO, DEBUG),
        (-10, WARNING, INFO),
        (-10, ERROR, WARNING),
        (-10, CRITICAL, ERROR),
    ],
)
def test_log_level_action(const, initial, expect):
    """``LogLevelAction`` increments/decrements the log level value correctly"""
    action = argparse_helpers.LogLevelAction([], "log_level", const)
    parser = ArgumentParser()
    namespace = Namespace()
    namespace.log_level = initial

    action(parser, namespace, [])

    assert namespace.log_level == expect


@pytest.mark.parametrize(
    "const, count, expect",
    [
        (10, NOTSET, WARNING),
        (10, 1, ERROR),
        (10, 2, CRITICAL),
        (10, 3, CRITICAL),
        (-10, NOTSET, WARNING),
        (-10, 1, INFO),
        (-10, 2, DEBUG),
        (-10, 3, DEBUG),
    ],
)
def test_argumentparser_log_level_action(const, count, expect):
    """The log level action works correctly with an ``ArgumentParser``"""
    parser = ArgumentParser()
    parser.register("action", "log_level", argparse_helpers.LogLevelAction)
    parser.add_argument("-l", dest="log_level", action="log_level", const=const)

    args = parser.parse_args(count * ["-l"])

    assert args.log_level == expect



"""Run arbitrary linters on given files in a Git repository

This supports any linter which reports on standard output and has a fairly standard
correct output format::

    <path>:<linenum>: <description>
    <path>:<linenum>:<column>: <description>

For example, Mypy outputs::

    module.py:57: error: Function is missing a type annotation for one or more arguments

Pylint, on the other hand::

    module.py:44:8: R1720: Unnecessary "elif" after "raise" (no-else-raise)

All such output from the linter will be printed on the standard output
provided that the ``<linenum>`` falls on a changed line.

"""

import logging
from pathlib import Path
from subprocess import PIPE, Popen
from typing import Set, Tuple, Union

from darker.git import EditedLinenumsDiffer

logger = logging.getLogger(__name__)


def _parse_linter_line(
    line: str, git_root: Path
) -> Union[Tuple[Path, int], Tuple[None, None]]:
    # Parse an error/note line.
    # Given: line == "dir/file.py:123: error: Foo\n"
    # Sets: path = Path("abs/path/to/dir/file.py:123"
    #       linenum = 123
    #       description = "error: Foo"
    try:
        location, _ = line[:-1].split(": ", 1)
        path_str, linenum_bytes, *rest = location.split(":")
        linenum = int(linenum_bytes)
        if len(rest) > 1:
            raise ValueError("Too many colon-separated tokens")
        if len(rest) == 1:
            # Make sure it column looks like an int on "<path>:<linenum>:<column>"
            _column = int(rest[0])  # noqa: F841
    except ValueError:
        # Encountered a non-parseable line which doesn't express a linting error.
        # For example, on Mypy:
        # "Found XX errors in YY files (checked ZZ source files)"
        # "Success: no issues found in 1 source file"
        logger.debug("Unparseable linter output: %s", line[:-1])
        return None, None
    path_from_cwd = Path(path_str).absolute()
    path_in_repo = path_from_cwd.relative_to(git_root)
    return path_in_repo, linenum


def run_linter(cmdline: str, git_root: Path, paths: Set[Path], revision: str) -> None:
    """Run the given linter and print linting errors falling on changed lines

    :param cmdline: The command line for running the linter
    :param git_root: The repository root for the changed files
    :param paths: Paths of files to check, relative to ``git_root``
    :param revision: The Git revision against which to compare the working tree

    """
    if not paths:
        return
    linter_process = Popen(
        cmdline.split() + [str(git_root / path) for path in sorted(paths)],
        stdout=PIPE,
        encoding="utf-8",
    )
    # assert needed for MyPy (see https://stackoverflow.com/q/57350490/15770)
    assert linter_process.stdout is not None
    edited_linenums_differ = EditedLinenumsDiffer(git_root, revision)
    for line in linter_process.stdout:
        path_in_repo, linter_error_linenum = _parse_linter_line(line, git_root)
        if path_in_repo is None:
            continue
        edited_linenums = edited_linenums_differ.revision_vs_worktree(
            path_in_repo, context_lines=0
        )
        if linter_error_linenum in edited_linenums:
            print(line, end="")



"""Miscellaneous utility functions"""

import io
from itertools import chain
from pathlib import Path
from typing import Iterable, List, Tuple, Union


def debug_dump(
    black_chunks: List[Tuple[int, List[str], List[str]]],
    old_content: str,
    new_content: str,
    edited_linenums: List[int],
) -> None:
    """Print debug output. This is used in case of an unexpected failure."""
    for offset, old_lines, new_lines in black_chunks:
        print(80 * "-")
        for delta, old_line in enumerate(old_lines):
            linenum = offset + delta
            edited = "*" if linenum in edited_linenums else " "
            print(f"{edited}-{linenum:4} {old_line}")
        for _, new_line in enumerate(new_lines):
            print(f" +     {new_line}")
    print(80 * "-")


def joinlines(lines: List[str]) -> str:
    """Join a list of lines back, adding a linefeed after each line

    This is the reverse of ``str.splitlines()``.

    """
    return "".join(f"{line}\n" for line in lines)


def get_path_ancestry(path: Path) -> Iterable[Path]:
    reverse_parents = reversed(path.parents)
    if path.is_dir():
        return chain(reverse_parents, [path])
    else:
        return reverse_parents


def get_common_root(paths: Iterable[Path]) -> Path:
    """Find the deepest common parent directory of given paths"""
    resolved_paths = [path.resolve() for path in paths]
    parents = reversed(list(zip(*(get_path_ancestry(path) for path in resolved_paths))))
    for first_path, *other_paths in parents:
        if all(path == first_path for path in other_paths):
            return first_path
    raise ValueError(f"Paths have no common parent Git root: {resolved_paths}")


class Buf:
    def __init__(self, initial_bytes: bytes):
        self._buf = io.BytesIO(initial_bytes)
        self._line_starts: List[int] = []

    def __next__(self) -> str:
        self._line_starts.append(self._buf.tell())
        return next(self._buf).rstrip(b"\n").decode("utf-8")

    def __iter__(self) -> "Buf":
        return self

    def seek_line(self, lines_delta: int) -> None:
        assert lines_delta <= 0
        for _ in range(-lines_delta):
            self._buf.seek(self._line_starts.pop())

    def next_line_startswith(self, prefix: Union[str, Tuple[str, ...]]) -> bool:
        try:
            return next(self).startswith(prefix)
        except StopIteration:
            return False
        finally:
            self.seek_line(-1)



from pathlib import Path
from unittest.mock import patch

import pytest

from darker.git import (
    EditedLinenumsDiffer,
    git_get_modified_files,
    git_get_unmodified_content,
    should_reformat_file,
)


@pytest.mark.parametrize(
    "revision, expect",
    [("HEAD", ["modified content"]), ("HEAD^", ["original content"]), ("HEAD~2", []),],
)
def test_git_get_unmodified_content(git_repo, revision, expect):
    git_repo.add({"my.txt": "original content"}, commit="Initial commit")
    paths = git_repo.add({"my.txt": "modified content"}, commit="Initial commit")
    paths['my.txt'].write('new content')

    original_content = git_get_unmodified_content(
        Path("my.txt"), revision, cwd=git_repo.root
    )

    assert original_content == expect


@pytest.mark.parametrize(
    "revision, expect",
    [
        ("HEAD^", "git show HEAD^:./my.txt"),
        ("master..HEAD", "git show master:./my.txt"),
        ("master...HEAD", "git show master:./my.txt"),
        ("master..", "git show master:./my.txt"),
        ("master...", "git show master:./my.txt"),
    ],
)
def test_git_get_unmodified_content_range(revision, expect):
    with patch("darker.git.check_output") as check_output:

        git_get_unmodified_content(Path("my.txt"), revision, Path("cwd"))

        check_output.assert_called_once_with(
            expect.split(), cwd="cwd", encoding="utf-8"
        )


@pytest.mark.parametrize(
    'path, create, expect',
    [
        ('.', False, False),
        ('main', True, False),
        ('main.c', True, False),
        ('main.py', True, True),
        ('main.py', False, False),
        ('main.pyx', True, False),
        ('main.pyi', True, False),
        ('main.pyc', True, False),
        ('main.pyo', True, False),
        ('main.js', True, False),
    ],
)
def test_should_reformat_file(tmpdir, path, create, expect):
    if create:
        (tmpdir / path).ensure()

    result = should_reformat_file(Path(tmpdir / path))

    assert result == expect


@pytest.mark.parametrize(
    'modify_paths, paths, expect',
    [
        ({}, ['a.py'], []),
        ({}, [], []),
        ({'a.py': 'new'}, [], ['a.py']),
        ({'a.py': 'new'}, ['b.py'], []),
        ({'a.py': 'new'}, ['a.py', 'b.py'], ['a.py']),
        ({'c/d.py': 'new'}, ['c/d.py', 'd/f/g.py'], ['c/d.py']),
        ({'c/e.js': 'new'}, ['c/e.js'], []),
        ({'a.py': 'original'}, ['a.py'], []),
        ({'a.py': None}, ['a.py'], []),
        ({"h.py": "untracked"}, ["h.py"], ["h.py"]),
        ({}, ["h.py"], []),
    ],
)
def test_git_get_modified_files(git_repo, modify_paths, paths, expect):
    """Tests for `darker.git.git_get_modified_files()`"""
    root = Path(git_repo.root)
    git_repo.add(
        {
            'a.py': 'original',
            'b.py': 'original',
            'c/d.py': 'original',
            'c/e.js': 'original',
            'd/f/g.py': 'original',
        },
        commit="Initial commit",
    )
    for path, content in modify_paths.items():
        absolute_path = git_repo.root / path
        if content is None:
            absolute_path.remove()
        else:
            absolute_path.write(content, ensure=True)
    result = git_get_modified_files({root / p for p in paths}, "HEAD", cwd=root)
    assert {str(p) for p in result} == set(expect)


edited_linenums_differ_cases = pytest.mark.parametrize(
    "context_lines, expect",
    [
        (0, [3, 7]),
        (1, [2, 3, 4, 6, 7, 8]),
        (2, [1, 2, 3, 4, 5, 6, 7, 8]),
        (3, [1, 2, 3, 4, 5, 6, 7, 8]),
    ],
)


@edited_linenums_differ_cases
def test_edited_linenums_differ_revision_vs_worktree(git_repo, context_lines, expect):
    """Tests for EditedLinenumsDiffer.revision_vs_worktree()"""
    paths = git_repo.add({"a.py": "1\n2\n3\n4\n5\n6\n7\n8\n"}, commit="Initial commit")
    paths["a.py"].write("1\n2\nthree\n4\n5\n6\nseven\n8\n")
    differ = EditedLinenumsDiffer(git_repo.root, "HEAD")

    result = differ.revision_vs_worktree(Path("a.py"), context_lines)

    assert result == expect


@edited_linenums_differ_cases
def test_edited_linenums_differ_revision_vs_lines(git_repo, context_lines, expect):
    """Tests for EditedLinenumsDiffer.revision_vs_lines()"""
    git_repo.add({'a.py': '1\n2\n3\n4\n5\n6\n7\n8\n'}, commit='Initial commit')
    lines = ['1', '2', 'three', '4', '5', '6', 'seven', '8']
    differ = EditedLinenumsDiffer(git_repo.root, "HEAD")

    result = differ.revision_vs_lines(Path("a.py"), lines, context_lines)

    assert result == expect



"""Custom formatter and action for argparse"""

import logging
import re
from argparse import Action, ArgumentParser, HelpFormatter, Namespace
from textwrap import fill
from typing import Any, List, Sequence, Union

WORD_RE = re.compile(r"\w")


def _fill_line(line: str, width: int, indent: str) -> str:
    first_word_match = WORD_RE.search(line)
    first_word_offset = first_word_match.start() if first_word_match else 0
    return fill(
        line,
        width,
        initial_indent=indent,
        subsequent_indent=indent + first_word_offset * " ",
    )


class NewlinePreservingFormatter(HelpFormatter):
    def _fill_text(self, text: str, width: int, indent: str) -> str:
        if "\n" in text:
            return "\n".join(
                _fill_line(line, width, indent) for line in text.split("\n")
            )
        return super()._fill_text(text, width, indent)


class LogLevelAction(Action):  # pylint: disable=too-few-public-methods
    """Support for command line actions which increment/decrement the log level"""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        option_strings: List[str],
        dest: str,
        const: int,
        default: int = logging.WARNING,
        required: bool = False,
        help: str = None,  # pylint: disable=redefined-builtin
        metavar: str = None,
    ):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            const=const,
            default=default,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: str = None,
    ) -> None:
        assert isinstance(values, list)
        assert all(isinstance(v, str) for v in values)
        current_level = getattr(namespace, self.dest, self.default)
        new_level = current_level + self.const
        new_level = max(new_level, logging.DEBUG)
        new_level = min(new_level, logging.CRITICAL)
        setattr(namespace, self.dest, new_level)



import pytest

from darker.verification import NotEquivalentError, verify_ast_unchanged


@pytest.mark.parametrize(
    "src_content, dst_content, expect",
    [
        ("if True: pass\n", "if False: pass\n", AssertionError),
        ("if True: pass\n", "if True:\n    pass\n", None),
    ],
)
def test_verify_ast_unchanged(src_content, dst_content, expect):
    black_chunks = [(1, ["black"], ["chunks"])]
    edited_linenums = [1, 2]
    try:
        verify_ast_unchanged(src_content, dst_content, black_chunks, edited_linenums)
    except NotEquivalentError:
        assert expect is AssertionError
    else:
        assert expect is None



import pytest

from darker.__main__ import main
from darker.tests.helpers import raises_if_exception

# The following test is a bit dense, so some explanation is due.
#
# A Git repository with 3 commits is created. Python files with one line of code
# ("ORIGINAL=1") are added, modified or deleted in each commit. An abbreviated number is
# used for each commit in the test code:
# - 2: HEAD~2          add 4 files
# - 1: HEAD~1 (HEAD^)  add 2 files, delete 1 file, keep 1 file, modify 2 files
# - 0: HEAD~0 (HEAD)   delete 1 file, modify 1 file, keep 4 files
#
# In each test case, all of the 6 files are overwritten in the working tree with given
# content.
#
# `darker --diff --revision <rev>` is called and the diff captured from stdout.
#
# Each case is gets as parameters:
# - the `<rev>` argument – which Git revision to compare to
# - `worktree_content` – the content to write to all the 6 known Python files
# - `expect` – a list of files expected to appear in `darker --diff`
#
# The Python file names tell the history of the file with a list of actions:
# +2 - added in HEAD~2
# +1 - added in HEAD^
# -1 - deleted in HEAD^
# M1 - modified in HEAD^
# -0 - deleted in HEAD
# M0 - modified in HEAD
#
# So e.g. the test case `HEAD~2`, `ORIGINAL=1` tells us that when `ORIGINAL=1` is
# written to all Python files and `darker --diff --revision HEAD~2` is called, the diff
# indicates reformatting in
# - +1.py – the file which was added in HEAD^
# - +1M0.py – the file added in HEAD^ and overwritten in HEAD (with `MODIFIED=1`)
# All other files are missing from the diff since their content was the same as now
# (`ORIGINAL=1`) at HEAD~2.


@pytest.mark.parametrize(
    "revision, worktree_content, expect",
    [
        ("", "USERMOD=1\n", ["+1", "+1M0", "+2-1", "+2", "+2M1-0", "+2M1"]),
        ("", "ORIGINAL=1\n", ["+1M0", "+2-1", "+2M1-0", "+2M1"]),
        ("", "MODIFIED=1\n", ["+1", "+2-1", "+2", "+2M1-0"]),
        ("HEAD", "USERMOD=1\n", ["+1", "+1M0", "+2-1", "+2", "+2M1-0", "+2M1"]),
        ("HEAD", "ORIGINAL=1\n", ["+1M0", "+2-1", "+2M1-0", "+2M1"],),
        ("HEAD", "MODIFIED=1\n", ["+1", "+2-1", "+2", "+2M1-0"]),
        ("HEAD^", "USERMOD=1\n", ["+1", "+1M0", "+2-1", "+2", "+2M1-0", "+2M1"]),
        ("HEAD^", "USERMOD=1\n", ["+1", "+1M0", "+2-1", "+2", "+2M1-0", "+2M1"]),
        ("HEAD^", "ORIGINAL=1\n", ["+2-1", "+2M1-0", "+2M1"]),
        ("HEAD^", "MODIFIED=1\n", ["+1", "+1M0", "+2-1", "+2"]),
        ("HEAD~2", "USERMOD=1\n", ["+1", "+1M0", "+2-1", "+2", "+2M1-0", "+2M1"]),
        ("HEAD~2", "ORIGINAL=1\n", ["+1", "+1M0"]),
        ("HEAD~2", "MODIFIED=1\n", ["+1", "+1M0", "+2-1", "+2", "+2M1-0", "+2M1"]),
        ("HEAD~3", "USERMOD=1\n", SystemExit),
    ],
)
def test_revision(git_repo, monkeypatch, capsys, revision, worktree_content, expect):
    monkeypatch.chdir(git_repo.root)
    # 2: HEAD~2:
    paths = git_repo.add(
        {
            "+2.py": "ORIGINAL=1\n",
            "+2M1.py": "ORIGINAL=1\n",
            "+2-1.py": "ORIGINAL=1\n",
            "+2M1-0.py": "ORIGINAL=1\n",
        },
        commit="First commit",
    )
    # 1: HEAD~1 i.e. HEAD^
    paths.update(
        git_repo.add(
            {
                "+2M1.py": "MODIFIED=1\n",
                "+1.py": "ORIGINAL=1\n",
                "+1M0.py": "ORIGINAL=1\n",
                "+2-1.py": None,
                "+2M1-0.py": "MODIFIED=1\n",
            },
            commit="Second commit",
        )
    )
    # 0: HEAD~0 i.e. HEAD:
    git_repo.add(
        {"+1M0.py": "MODIFIED=1\n", "+2M1-0.py": None}, commit="Third commit",
    )
    # Working tree:
    for path in paths.values():
        path.write(worktree_content)
    arguments = ["--diff", "--revision", revision, "."]

    with raises_if_exception(expect):
        main(arguments)

        modified_files = [
            line[4:-3]
            for line in capsys.readouterr().out.splitlines()
            if line.startswith("+++ ")
        ]
        assert modified_files == expect



import logging
import sys
from pathlib import Path
from typing import Optional

from black import find_project_root

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


try:
    import isort
except ImportError:
    isort = None

logger = logging.getLogger(__name__)


class IsortArgs(TypedDict, total=False):
    line_length: int
    settings_file: str
    settings_path: str


def apply_isort(
    content: str,
    src: Path,
    config: Optional[str] = None,
    line_length: Optional[int] = None,
) -> str:
    isort_args = IsortArgs()
    if config:
        isort_args["settings_file"] = config
    else:
        isort_args["settings_path"] = str(find_project_root((str(src),)))
    if line_length:
        isort_args["line_length"] = line_length

    logger.debug(
        "isort.code(code=..., {})".format(
            ", ".join(f"{k}={v!r}" for k, v in isort_args.items())
        )
    )
    result: str = isort.code(code=content, **isort_args)
    return result



from textwrap import dedent

ORIGINAL = dedent(
    '''\
    original first line
    original second line
    original third line
'''
)

CHANGED = dedent(
    '''\
    original first line
    changed second line
    original third line
'''
)



from difflib import SequenceMatcher

from darker.tests.git_diff_example_output import CHANGED, ORIGINAL


def test_sequencematcher():
    s = SequenceMatcher(None, ORIGINAL.splitlines(), CHANGED.splitlines())
    assert s.get_opcodes() == [
        ("equal", 0, 1, 0, 1),
        ("replace", 1, 2, 1, 2),
        ("equal", 2, 3, 2, 3),
    ]



from pathlib import Path
from subprocess import check_call
from types import SimpleNamespace
from unittest.mock import Mock, patch

import pytest
from black import find_project_root

import darker.__main__
import darker.import_sorting


def test_isort_option_without_isort(tmpdir, without_isort, caplog):
    check_call(["git", "init"], cwd=tmpdir)
    with patch.object(darker.__main__, "isort", None), pytest.raises(SystemExit):

        darker.__main__.main(["--isort", str(tmpdir)])

    assert (
        "Please run `pip install 'darker[isort]'` to use the `--isort` option."
        in caplog.text
    )


@pytest.fixture
def run_isort(git_repo, monkeypatch, caplog, request):
    find_project_root.cache_clear()

    monkeypatch.chdir(git_repo.root)
    paths = git_repo.add({'test1.py': 'original'}, commit='Initial commit')
    paths['test1.py'].write('changed')
    args = getattr(request, "param", ())
    with patch.multiple(
        darker.__main__, run_black=Mock(return_value=[]), verify_ast_unchanged=Mock(),
    ), patch("darker.import_sorting.isort.code"):
        darker.__main__.main(["--isort", "./test1.py", *args])
        return SimpleNamespace(
            isort_code=darker.import_sorting.isort.code, caplog=caplog
        )


def test_isort_option_with_isort(run_isort):
    assert "Please run" not in run_isort.caplog.text


@pytest.mark.parametrize(
    "run_isort, isort_args",
    [((), {}), (("--line-length", "120"), {"line_length": 120})],
    indirect=["run_isort"],
)
def test_isort_option_with_isort_calls_sortimports(tmpdir, run_isort, isort_args):
    run_isort.isort_code.assert_called_once_with(
        code="changed", settings_path=str(tmpdir), **isort_args
    )


def test_format_edited_parts_empty():
    with pytest.raises(ValueError):

        list(darker.__main__.format_edited_parts([], "HEAD", False, [], {}))


A_PY = ['import sys', 'import os', "print( '42')", '']
A_PY_BLACK = ['import sys', 'import os', '', 'print("42")', '']
A_PY_BLACK_UNNORMALIZE = ['import sys', 'import os', '', "print('42')", '']
A_PY_BLACK_ISORT = ['import os', 'import sys', '', 'print("42")', '']

A_PY_DIFF_BLACK = [
    '--- a.py',
    '+++ a.py',
    '@@ -1,3 +1,4 @@',
    ' import sys',
    ' import os',
    "-print( '42')",
    '+',
    '+print("42")',
    '',
]

A_PY_DIFF_BLACK_NO_STR_NORMALIZE = [
    '--- a.py',
    '+++ a.py',
    '@@ -1,3 +1,4 @@',
    ' import sys',
    ' import os',
    "-print( '42')",
    '+',
    "+print('42')",
    '',
]

A_PY_DIFF_BLACK_ISORT = [
    '--- a.py',
    '+++ a.py',
    '@@ -1,3 +1,4 @@',
    '+import os',
    ' import sys',
    '-import os',
    "-print( '42')",
    '+',
    '+print("42")',
    '',
]


@pytest.mark.parametrize(
    'enable_isort, black_args, expect',
    [
        (False, {}, A_PY_BLACK),
        (True, {}, A_PY_BLACK_ISORT),
        (False, {'skip_string_normalization': True}, A_PY_BLACK_UNNORMALIZE),
    ],
)
def test_format_edited_parts(git_repo, monkeypatch, enable_isort, black_args, expect):
    monkeypatch.chdir(git_repo.root)
    paths = git_repo.add({'a.py': '\n', 'b.py': '\n'}, commit='Initial commit')
    paths['a.py'].write('\n'.join(A_PY))
    paths['b.py'].write('print(42 )\n')

    changes = list(
        darker.__main__.format_edited_parts(
            [Path("a.py")], "HEAD", enable_isort, [], black_args
        )
    )

    expect_changes = [(paths['a.py'], '\n'.join(A_PY), '\n'.join(expect), expect[:-1])]
    assert changes == expect_changes


def test_format_edited_parts_all_unchanged(git_repo, monkeypatch):
    """``format_edited_parts()`` yields nothing if no reformatting was needed"""
    monkeypatch.chdir(git_repo.root)
    paths = git_repo.add({'a.py': 'pass\n', 'b.py': 'pass\n'}, commit='Initial commit')
    paths['a.py'].write('"properly"\n"formatted"\n')
    paths['b.py'].write('"not"\n"checked"\n')

    result = list(
        darker.__main__.format_edited_parts([Path("a.py")], "HEAD", True, [], {})
    )

    assert result == []


@pytest.mark.parametrize(
    'arguments, expect_stdout, expect_a_py, expect_retval',
    [
        (['--diff'], A_PY_DIFF_BLACK, A_PY, 0),
        (['--isort'], [''], A_PY_BLACK_ISORT, 0),
        (
            ['--skip-string-normalization', '--diff'],
            A_PY_DIFF_BLACK_NO_STR_NORMALIZE,
            A_PY,
            0,
        ),
        ([], [''], A_PY_BLACK, 0),
        (['--isort', '--diff'], A_PY_DIFF_BLACK_ISORT, A_PY, 0),
        (['--check'], [''], A_PY, 1),
        (['--check', '--diff'], A_PY_DIFF_BLACK, A_PY, 1),
        (['--check', '--isort'], [''], A_PY, 1),
        (['--check', '--diff', '--isort'], A_PY_DIFF_BLACK_ISORT, A_PY, 1),
    ],
)
def test_main(
    git_repo, monkeypatch, capsys, arguments, expect_stdout, expect_a_py, expect_retval
):  # pylint: disable=too-many-arguments
    """Main function outputs diffs and modifies files correctly"""
    monkeypatch.chdir(git_repo.root)
    paths = git_repo.add({'a.py': '\n', 'b.py': '\n'}, commit='Initial commit')
    paths['a.py'].write('\n'.join(A_PY))
    paths['b.py'].write('print(42 )\n')

    retval = darker.__main__.main(arguments + ['a.py'])

    stdout = capsys.readouterr().out.replace(str(git_repo.root), '')
    assert stdout.split('\n') == expect_stdout
    assert paths['a.py'].readlines(cr=False) == expect_a_py
    assert paths['b.py'].readlines(cr=False) == ['print(42 )', '']
    assert retval == expect_retval


def test_output_diff(capsys):
    """output_diff() prints Black-style diff output"""
    darker.__main__.print_diff(
        Path('a.py'),
        'unchanged\nremoved\nkept 1\n2\n3\n4\n5\n6\n7\nchanged\n',
        ['inserted', 'unchanged', 'kept 1', '2', '3', '4', '5', '6', '7', 'Changed'],
    )

    assert capsys.readouterr().out.splitlines() == [
        '--- a.py',
        '+++ a.py',
        '@@ -1,5 +1,5 @@',
        '+inserted',
        ' unchanged',
        '-removed',
        ' kept 1',
        ' 2',
        ' 3',
        '@@ -7,4 +7,4 @@',
        ' 5',
        ' 6',
        ' 7',
        '-changed',
        '+Changed',
    ]



import re

from setuptools import setup

CONTRIBUTORS_RE = re.compile(r"""\nThanks goes .*\nThis project follows """, re.DOTALL)


def make_pypi_compliant_readme() -> str:
    """Remove raw HTML from ``README.rst`` before packaging

    The contributors table should only be shown on GitHub. It must be removed before
    being displayed on PyPI. The simplest way to do this is to simply match and strip it
    when creating a distribution archive.

    This function reads the contents of ``README.rst``, replaces the row HTML section
    with a plain-text link to the README on GitHub and returns the resulting string.

    :return: The contents of a PyPI compliant ``README.rst``

    """
    with open("README.rst") as fp:
        original_readme = fp.read()
    modified_readme = CONTRIBUTORS_RE.sub(
        "\nSee README.rst_ for the list of contributors.\n\nThis project follows ",
        original_readme,
    )
    if modified_readme == original_readme:
        raise RuntimeError(
            f"The contributors table couldn't be found in README.rst using the pattern "
            f"'{CONTRIBUTORS_RE.pattern}'"
        )
    return modified_readme


setup(long_description=make_pypi_compliant_readme())



from argparse import ArgumentParser, Namespace
from textwrap import dedent

import pytest

from darker.config import (
    TomlArrayLinesEncoder,
    dump_config,
    get_effective_config,
    get_modified_config,
    load_config,
    replace_log_level_name,
)


@pytest.mark.parametrize(
    "list_value, expect",
    [
        ([], "[\n]"),
        (["one value"], '[\n    "one value",\n]'),
        (["two", "values"], '[\n    "two",\n    "values",\n]'),
        (
            [
                "a",
                "dozen",
                "short",
                "string",
                "values",
                "in",
                "the",
                "list",
                "of",
                "strings",
                "to",
                "format",
            ],
            '[\n    "a",\n    "dozen",\n    "short",\n    "string",\n    "values"'
            ',\n    "in",\n    "the",\n    "list",\n    "of",\n    "strings"'
            ',\n    "to",\n    "format",\n]',
        ),
    ],
)
def test_toml_array_lines_encoder(list_value, expect):
    result = TomlArrayLinesEncoder().dump_list(list_value)

    assert result == expect


@pytest.mark.parametrize(
    "log_level, expect",
    [
        (0, "NOTSET"),
        (10, "DEBUG"),
        (20, "INFO"),
        (30, "WARNING"),
        (40, "ERROR"),
        (50, "CRITICAL"),
        ("DEBUG", 10),
        ("INFO", 20),
        ("WARNING", 30),
        ("WARN", 30),
        ("ERROR", 40),
        ("CRITICAL", 50),
        ("FOOBAR", "Level FOOBAR"),
    ],
)
def test_replace_log_level_name(log_level, expect):
    config = {} if log_level is None else {"log_level": log_level}
    replace_log_level_name(config)

    assert config["log_level"] == expect


@pytest.mark.parametrize(
    "srcs, cwd, expect",
    [
        ([], ".", {"CONFIG_PATH": "."}),
        ([], "level1", {"CONFIG_PATH": "."}),
        ([], "level1/level2", {"CONFIG_PATH": "."}),
        ([], "has_git", {}),
        ([], "has_git/level1", {}),
        ([], "has_pyproject", {"CONFIG_PATH": "has_pyproject"}),
        ([], "has_pyproject/level1", {"CONFIG_PATH": "has_pyproject"}),
        (["root.py"], ".", {"CONFIG_PATH": "."}),
        (["../root.py"], "level1", {"CONFIG_PATH": "."}),
        (["../root.py"], "has_git", {"CONFIG_PATH": "."}),
        (["../root.py"], "has_pyproject", {"CONFIG_PATH": "."}),
        (["root.py", "level1/level1.py"], ".", {"CONFIG_PATH": "."}),
        (["../root.py", "level1.py"], "level1", {"CONFIG_PATH": "."}),
        (["../root.py", "../level1/level1.py"], "has_git", {"CONFIG_PATH": "."}),
        (["../root.py", "../level1/level1.py"], "has_pyproject", {"CONFIG_PATH": "."}),
        (["has_pyproject/pyp.py", "level1/level1.py"], ".", {"CONFIG_PATH": "."}),
        (["../has_pyproject/pyp.py", "level1.py"], "level1", {"CONFIG_PATH": "."}),
        (
            ["../has_pyproject/pyp.py", "../level1/level1.py"],
            "has_git",
            {"CONFIG_PATH": "."},
        ),
        (["pyp.py", "../level1/level1.py"], "has_pyproject", {"CONFIG_PATH": "."}),
        (
            ["has_pyproject/level1/l1.py", "has_pyproject/level1b/l1b.py"],
            ".",
            {"CONFIG_PATH": "has_pyproject"},
        ),
        (
            ["../has_pyproject/level1/l1.py", "../has_pyproject/level1b/l1b.py"],
            "level1",
            {"CONFIG_PATH": "has_pyproject"},
        ),
        (
            ["../has_pyproject/level1/l1.py", "../has_pyproject/level1b/l1b.py"],
            "has_git",
            {"CONFIG_PATH": "has_pyproject"},
        ),
        (
            ["level1/l1.py", "level1b/l1b.py"],
            "has_pyproject",
            {"CONFIG_PATH": "has_pyproject"},
        ),
        (
            ["full_example/full.py"],
            ".",
            {
                "check": True,
                "diff": True,
                "isort": True,
                "lint": ["flake8", "mypy", "pylint"],
                "log_level": 10,
                "revision": "main",
                "src": ["src", "tests"],
            },
        ),
    ],
)
def test_load_config(
    find_project_root_cache_clear, tmp_path, monkeypatch, srcs, cwd, expect
):
    (tmp_path / ".git").mkdir()
    (tmp_path / "pyproject.toml").write_text('[tool.darker]\nCONFIG_PATH = "."\n')
    (tmp_path / "level1/level2").mkdir(parents=True)
    (tmp_path / "has_git/.git").mkdir(parents=True)
    (tmp_path / "has_git/level1").mkdir()
    (tmp_path / "has_pyproject/level1").mkdir(parents=True)
    (tmp_path / "has_pyproject/pyproject.toml").write_text(
        '[tool.darker]\nCONFIG_PATH = "has_pyproject"\n'
    )
    (tmp_path / "full_example").mkdir()
    (tmp_path / "full_example/pyproject.toml").write_text(
        dedent(
            """
            [tool.darker]
            src = [
                "src",
                "tests",
            ]
            revision = "main"
            diff = true
            check = true
            isort = true
            lint = [
                "flake8",
                "mypy",
                "pylint",
            ]
            log_level = "DEBUG"
            """
        )
    )
    monkeypatch.chdir(tmp_path / cwd)

    result = load_config(srcs)

    assert result == expect


@pytest.mark.parametrize(
    "args, expect",
    [
        (Namespace(), {}),
        (Namespace(one="option"), {"one": "option"}),
        (Namespace(log_level=10), {"log_level": "DEBUG"}),
        (
            Namespace(two="options", log_level=20),
            {"two": "options", "log_level": "INFO"},
        ),
    ],
)
def test_get_effective_config(args, expect):
    result = get_effective_config(args)

    assert result == expect


@pytest.mark.parametrize(
    "args, expect",
    [
        (Namespace(), {}),
        (Namespace(unknown="option"), {"unknown": "option"}),
        (Namespace(log_level=10), {"log_level": "DEBUG"}),
        (Namespace(names=[], int=42, string="fourty-two"), {"names": []}),
        (Namespace(names=["bar"], int=42, string="fourty-two"), {"names": ["bar"]}),
        (Namespace(names=["foo"], int=43, string="fourty-two"), {"int": 43}),
        (Namespace(names=["foo"], int=42, string="one"), {"string": "one"}),
    ],
)
def test_get_modified_config(args, expect):
    parser = ArgumentParser()
    parser.add_argument("names", nargs="*", default=["foo"])
    parser.add_argument("--int", dest="int", default=42)
    parser.add_argument("--string", default="fourty-two")
    result = get_modified_config(parser, args)

    assert result == expect


@pytest.mark.parametrize(
    "config, expect",
    [
        ({}, "[tool.darker]\n"),
        ({"str": "value"}, '[tool.darker]\nstr = "value"\n'),
        ({"int": 42}, "[tool.darker]\nint = 42\n"),
        ({"float": 4.2}, "[tool.darker]\nfloat = 4.2\n"),
        (
            {"list": ["foo", "bar"]},
            dedent(
                """\
                [tool.darker]
                list = [
                    "foo",
                    "bar",
                ]
                """
            ),
        ),
        (
            {
                "src": ["main.py"],
                "revision": "master",
                "diff": False,
                "check": False,
                "isort": False,
                "lint": [],
                "config": None,
                "log_level": "DEBUG",
                "skip_string_normalization": None,
                "line_length": None,
            },
            dedent(
                """\
                [tool.darker]
                src = [
                    "main.py",
                ]
                revision = "master"
                diff = false
                check = false
                isort = false
                lint = [
                ]
                log_level = "DEBUG"
                """
            ),
        ),
    ],
)
def test_dump_config(config, expect):
    result = dump_config(config)

    assert result == expect



"""Load and save configuration in TOML format"""

import logging
from argparse import ArgumentParser, Namespace
from typing import Dict, Iterable, List, Union, cast

import toml
from black import find_project_root


class TomlArrayLinesEncoder(toml.TomlEncoder):  # type: ignore[name-defined]
    """Format TOML so list items are each on their own line"""

    def dump_list(self, v: List[str]) -> str:
        """Format a list value"""
        return "[{}\n]".format(
            "".join("\n    {},".format(self.dump_value(item)) for item in v)
        )


DarkerConfig = Dict[str, Union[str, bool, List[str]]]


def replace_log_level_name(config: DarkerConfig) -> None:
    """Replace numeric log level in configuration with the name of the log level"""
    if "log_level" in config:
        config["log_level"] = logging.getLevelName(cast(int, config["log_level"]))


def load_config(srcs: Iterable[str]) -> DarkerConfig:
    """Find and load Darker configuration from given path or pyproject.toml

    :param srcs: File(s) and directory/directories which will be processed. Their paths
                 are used to look for the ``pyproject.toml`` configuration file.

    """
    path = find_project_root(tuple(srcs or ["."])) / "pyproject.toml"
    if path.is_file():
        pyproject_toml = toml.load(path)
        config: DarkerConfig = pyproject_toml.get("tool", {}).get("darker", {}) or {}
        replace_log_level_name(config)
        return config
    return {}


def get_effective_config(args: Namespace) -> DarkerConfig:
    """Return all configuration options"""
    config = vars(args).copy()
    replace_log_level_name(config)
    return config


def get_modified_config(parser: ArgumentParser, args: Namespace) -> DarkerConfig:
    """Return configuration options which are set to non-default values"""
    not_default = {
        argument: value
        for argument, value in vars(args).items()
        if value != parser.get_default(argument)
    }
    replace_log_level_name(not_default)
    return not_default


def dump_config(config: DarkerConfig) -> str:
    """Return the configuration in TOML format"""
    dump = toml.dumps(config, encoder=TomlArrayLinesEncoder())  # type: ignore[call-arg]
    return f"[tool.darker]\n{dump}"



from itertools import chain
from textwrap import dedent

import pytest

from darker.diff import (
    diff_and_get_opcodes,
    opcodes_to_chunks,
    opcodes_to_edit_linenums,
)

FUNCTIONS2_PY = dedent(
    """\
    def f(
      a,
      **kwargs,
    ) -> A:
        with cache_dir():
            if something:
                result = (
                    CliRunner().invoke(black.main, [str(src1), str(src2), "--diff", "--check"])
                )
        limited.append(-limited.pop())  # negate top
        return A(
            very_long_argument_name1=very_long_value_for_the_argument,
            very_long_argument_name2=-very.long.value.for_the_argument,
            **kwargs,
        )
    def g():
        "Docstring."
        def inner():
            pass
        print("Inner defs should breathe a little.")
    def h():
        def inner():
            pass
        print("Inner defs should breathe a little.")
"""  # noqa: E501
)


FUNCTIONS2_PY_REFORMATTED = dedent(
    """\
    def f(
        a,
        **kwargs,
    ) -> A:
        with cache_dir():
            if something:
                result = CliRunner().invoke(
                    black.main, [str(src1), str(src2), "--diff", "--check"]
                )
        limited.append(-limited.pop())  # negate top
        return A(
            very_long_argument_name1=very_long_value_for_the_argument,
            very_long_argument_name2=-very.long.value.for_the_argument,
            **kwargs,
        )


    def g():
        "Docstring."

        def inner():
            pass

        print("Inner defs should breathe a little.")


    def h():
        def inner():
            pass

        print("Inner defs should breathe a little.")
    """
)


EXPECT_OPCODES = [
    ("equal", 0, 1, 0, 1),
    ("replace", 1, 3, 1, 3),
    ("equal", 3, 6, 3, 6),
    ("replace", 6, 8, 6, 8),
    ("equal", 8, 15, 8, 15),
    ("insert", 15, 15, 15, 17),
    ("equal", 15, 17, 17, 19),
    ("insert", 17, 17, 19, 20),
    ("equal", 17, 19, 20, 22),
    ("insert", 19, 19, 22, 23),
    ("equal", 19, 20, 23, 24),
    ("insert", 20, 20, 24, 26),
    ("equal", 20, 23, 26, 29),
    ("insert", 23, 23, 29, 30),
    ("equal", 23, 24, 30, 31),
]


def test_diff_and_get_opcodes():
    src_lines = FUNCTIONS2_PY.splitlines()
    dst_lines = FUNCTIONS2_PY_REFORMATTED.splitlines()
    opcodes = diff_and_get_opcodes(src_lines, dst_lines)
    assert opcodes == EXPECT_OPCODES


def test_opcodes_to_chunks():
    src_lines = FUNCTIONS2_PY.splitlines()
    dst_lines = FUNCTIONS2_PY_REFORMATTED.splitlines()

    chunks = list(opcodes_to_chunks(EXPECT_OPCODES, src_lines, dst_lines))

    assert chunks == [
        (1, ["def f("], ["def f("]),
        (2, ["  a,", "  **kwargs,"], ["    a,", "    **kwargs,"]),
        (
            4,
            [") -> A:", "    with cache_dir():", "        if something:"],
            [") -> A:", "    with cache_dir():", "        if something:"],
        ),
        (
            7,
            [
                "            result = (",
                "                CliRunner().invoke(black.main, [str(src1), str(src2), "
                '"--diff", "--check"])',
            ],
            [
                "            result = CliRunner().invoke(",
                '                black.main, [str(src1), str(src2), "--diff", "--check"]',  # noqa: E501
            ],
        ),
        (
            9,
            [
                "            )",
                "    limited.append(-limited.pop())  # negate top",
                "    return A(",
                "        very_long_argument_name1=very_long_value_for_the_argument,",
                "        very_long_argument_name2=-very.long.value.for_the_argument,",
                "        **kwargs,",
                "    )",
            ],
            [
                "            )",
                "    limited.append(-limited.pop())  # negate top",
                "    return A(",
                "        very_long_argument_name1=very_long_value_for_the_argument,",
                "        very_long_argument_name2=-very.long.value.for_the_argument,",
                "        **kwargs,",
                "    )",
            ],
        ),
        (16, [], ["", ""]),
        (16, ["def g():", '    "Docstring."'], ["def g():", '    "Docstring."']),
        (18, [], [""]),
        (
            18,
            ["    def inner():", "        pass"],
            ["    def inner():", "        pass"],
        ),
        (20, [], [""]),
        (
            20,
            ['    print("Inner defs should breathe a little.")'],
            ['    print("Inner defs should breathe a little.")'],
        ),
        (21, [], ["", ""]),
        (
            21,
            ["def h():", "    def inner():", "        pass"],
            ["def h():", "    def inner():", "        pass"],
        ),
        (24, [], [""]),
        (
            24,
            ['    print("Inner defs should breathe a little.")'],
            ['    print("Inner defs should breathe a little.")'],
        ),
    ]


EXAMPLE_OPCODES = [
    ("replace", 0, 4, 0, 1),
    ("equal", 4, 6, 1, 3),
    ("replace", 6, 8, 3, 5),
    ("equal", 8, 15, 5, 12),
    ("insert", 15, 15, 12, 14),
    ("equal", 15, 17, 14, 16),
    ("insert", 17, 17, 16, 17),
    ("equal", 17, 19, 17, 19),
    ("insert", 19, 19, 19, 20),
    ("equal", 19, 20, 20, 21),
    ("insert", 20, 20, 21, 23),
    ("equal", 20, 23, 23, 26),
    ("insert", 23, 23, 26, 27),
    ("equal", 23, 24, 27, 28),
]


@pytest.mark.parametrize(
    'context_lines, expect',
    [
        (0, [1, 4, 5, 13, 14, 17, 20, 22, 23, 27]),
        (1, [[1, 6], [12, 24], [26, 28]]),
        (2, [[1, 7], [11, 28]]),
    ],
)
def test_opcodes_to_edit_linenums(context_lines, expect):
    edit_linenums = list(opcodes_to_edit_linenums(EXAMPLE_OPCODES, context_lines))
    expect_ranges = [[n, n] if isinstance(n, int) else n for n in expect]
    expect_linenums = list(chain(*(range(n[0], n[1] + 1) for n in expect_ranges)))

    assert edit_linenums == expect_linenums


def test_opcodes_to_edit_linenums_empty_opcodes():
    result = list(opcodes_to_edit_linenums([], context_lines=0))

    assert result == []



from pathlib import Path
from textwrap import dedent

import pytest
from black import find_project_root

from darker.import_sorting import apply_isort

ORIGINAL_SOURCE = "import sys\nimport os\n"
ISORTED_SOURCE = "import os\nimport sys\n"


def test_apply_isort():
    result = apply_isort(ORIGINAL_SOURCE, Path("test1.py"))

    assert result == ISORTED_SOURCE


@pytest.mark.parametrize(
    "line_length, settings_file, expect",
    [
        (
            50,
            None,
            "from module import (ab, cd, ef, gh, ij, kl, mn,\n"
            "                    op, qr, st, uv, wx, yz)\n",
        ),
        (
            50,
            "pyproject.toml",
            "from module import (ab, cd, ef, gh, ij, kl, mn,\n"
            "                    op, qr, st, uv, wx, yz)\n",
        ),
        (
            60,
            None,
            "from module import (ab, cd, ef, gh, ij, kl, mn, op, qr, st,\n"
            "                    uv, wx, yz)\n",
        ),
        (
            60,
            "pyproject.toml",
            "from module import (ab, cd, ef, gh, ij, kl, mn, op, qr, st,\n"
            "                    uv, wx, yz)\n",
        ),
    ],
)
def test_isort_config(monkeypatch, tmpdir, line_length, settings_file, expect):
    find_project_root.cache_clear()
    monkeypatch.chdir(tmpdir)
    (tmpdir / 'pyproject.toml').write(
        dedent(
            f"""\
            [tool.isort]
            line_length = {line_length}
            """
        )
    )

    content = "from module import ab, cd, ef, gh, ij, kl, mn, op, qr, st, uv, wx, yz"
    config = str(tmpdir / settings_file) if settings_file else None

    actual = apply_isort(content, Path("test1.py"), config)
    assert actual == expect



import pytest

from darker.chooser import choose_lines


@pytest.mark.parametrize(
    "edited_line_numbers, expect",
    [
        ([], ["original first line", "original second line", "original third line"],),
        ([0], ["original first line", "original second line", "original third line"],),
        ([1], ["original first line", "changed second line", "original third line"],),
        ([2], ["original first line", "original second line", "original third line"],),
        (
            [0, 1],
            ["original first line", "changed second line", "original third line"],
        ),
    ],
)
def test_choose_edited_lines(edited_line_numbers, expect):
    black_chunks = [
        (0, ["original first line"], ["original first line"]),
        (1, ["original second line"], ["changed second line"]),
        (2, ["original third line"], ["original third line"]),
    ]
    result = list(choose_lines(black_chunks, edited_line_numbers))
    assert result == expect
